package de.contecon.picapport.plugin.dcraw;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Map.Entry;
import java.util.Properties;

import com.adobe.xmp.XMPException;

import de.contecon.imageutils.CcXMPMetaData;
import de.contecon.picapport.plugins.IPicApportPlugInLogger;
import de.contecon.picapport.plugins.otherformats.IOtherFileFormat;
import de.contecon.picapport.plugins.otherformats.OtherFormatsDescriptor;


/*
 * Created on 14.11.2014
 *
 * Author:      ES
 * Copyright:   Contecon Software GmbH
 * 
 * History:     14.11.2014 created
 * 
 */
public class DcRawPlugin implements IOtherFileFormat {

  private final static DateFormat EXIF_DATE_FORMAT = new SimpleDateFormat("yyyy:MM:dd HH:mm:ss");
  private final static DateFormat CTIME_DATE_FORMAT = new SimpleDateFormat("EEE MMM dd HH:mm:ss yyyy", Locale.US);
  
  private IPicApportPlugInLogger logger;
  private File pluginDirectory;
  
  private boolean preferThumbnails=false;         // If True Thumbnails will be taken from RAW rather than created from the RAW data
                                                  // This is usually the fastest way to get a jpg from a RAW but the orientation
                                                  // is sometimes wrong (depends on camera model)
  
  private String defaultConvertParams="-c -w -h"; // parameters passed to dcraw for conversation 
  
  private boolean isWindows=false; 
  
  private List<String> dcRawInfoCmd=null;
  private List<String> dcRawConvertCmd=null;
  private List<String> cjpegCmd=null;
  
  private List<String> dcRawThumbCmd=null;
  private int dcRawThumbCmdFilePos=0;
  
  
   
  
  @Override
  public List<OtherFormatsDescriptor> init(File pluginDirectory, Properties props, IPicApportPlugInLogger logger) {
    this.pluginDirectory = pluginDirectory;
    this.logger = logger;
    this.isWindows=System.getProperty("os.name").startsWith("Windows");
    
    String pt = props.getProperty("dcraw.prefer.thumbs", "false");
    if(pt.toLowerCase().trim().equals("true")) {
      this.preferThumbnails=true;
    }
    
    String dcp = props.getProperty("dcraw.convert.params", defaultConvertParams);
    String[] params = dcp.split("\\s+");
    
    if(isWindows) {
      dcRawInfoCmd=Arrays.asList(new String[]{pluginDirectory.getAbsolutePath()+File.separator+"dcraw.exe","-i","-v","rawfilename"});
      dcRawThumbCmd=Arrays.asList(new String[]{pluginDirectory.getAbsolutePath()+File.separator+"dcraw.exe","-e","rawfilename"});
      cjpegCmd=Arrays.asList(new String[]{pluginDirectory.getAbsolutePath()+File.separator+"cjpeg.exe","ppmfilename"});
      
      dcRawConvertCmd=new ArrayList<String>();
      dcRawConvertCmd.add(pluginDirectory.getAbsolutePath()+File.separator+"dcraw.exe");
      dcRawConvertCmd.addAll(Arrays.asList(params));
      dcRawConvertCmd.add("rawfilename");
      } else {
      dcRawInfoCmd=Arrays.asList(new String[]{"dcraw","-i","-v","rawfilename"});
      dcRawThumbCmd=Arrays.asList(new String[]{"dcraw","-e","rawfilename"});
      cjpegCmd=Arrays.asList(new String[]{"cjpeg","ppmfilename"});
      
      dcRawConvertCmd=new ArrayList<String>();
      dcRawConvertCmd.add("dcraw");
      dcRawConvertCmd.addAll(Arrays.asList(params));
      dcRawConvertCmd.add("rawfilename");
      }
    this.dcRawThumbCmdFilePos=dcRawConvertCmd.size()-1;
   
    ArrayList<OtherFormatsDescriptor> formats = new ArrayList<OtherFormatsDescriptor>();
    scanProps(props, formats);      
   
    return  formats;
}
  
  
  private void scanProps(Properties props, ArrayList<OtherFormatsDescriptor> formats) {
       for(Entry<Object, Object> e : props.entrySet()) {
        String key   = (String) e.getKey();
        if(key.startsWith(".") && key.endsWith(".name") && key.length() > 6) {
           String name= (String) e.getValue();
           String ext=  key.substring(0, key.length()-5);
           String mime= props.getProperty(ext+".mime", "");
           
           if(mime.length() > 0) {
             formats.add(
              new OtherFormatsDescriptor(ext, 
                  mime, 
                  true, 
                  "PicApport "+name+" plugin", 
                  "(c) 2015 Contecon Software GmbH", 
                  "1.0", 
                  props));
           }
        }
       }
    }
   

  @Override
  public CcXMPMetaData createJpegFile(File otherFormatFile, File jpegFileToCreate, CcXMPMetaData metaDataIn) throws Exception {
    String errorMessage = null;
    String metadata = null;
    
    try {
        File jpgThumbNail=getThumbFile(otherFormatFile, ".jpg");
        File ppmThumbNail=getThumbFile(otherFormatFile, ".ppm");
        
        dcRawInfoCmd.set(3, otherFormatFile.getAbsolutePath());
        dcRawThumbCmd.set(2, otherFormatFile.getAbsolutePath());
        
        metadata = new DcRawProcessWrapper(logger, this.pluginDirectory, dcRawInfoCmd).getSwStdOut();

        if(this.preferThumbnails) {
          new DcRawProcessWrapper(logger, this.pluginDirectory, dcRawThumbCmd);
          if(jpgThumbNail.exists()) {
            // Thumbnail was supported
            jpegFileToCreate.delete();
            jpgThumbNail.renameTo(jpegFileToCreate);
            } else {
            if(ppmThumbNail.exists()) {
              cjpegCmd.set(1, ppmThumbNail.getAbsolutePath());  
              new DcRawProcessWrapper(logger, this.pluginDirectory, cjpegCmd, jpegFileToCreate);
              ppmThumbNail.delete();
              } else {
              // renderRaw File
              errorMessage = renderRaw(otherFormatFile, jpegFileToCreate, ppmThumbNail);
              }  
            }
        } else {
         // renderRaw File
        errorMessage = renderRaw(otherFormatFile, jpegFileToCreate, ppmThumbNail);  
        }
        scanMetaData(otherFormatFile, metadata, metaDataIn);
    }
    catch (Exception e) {
      errorMessage = e.getLocalizedMessage();
    }

    if (errorMessage != null) {
      logger.LogErrorMessage("Unecpected Exception in createJpegFile: " + errorMessage);
      copyDummySubstitute(jpegFileToCreate, "RAW-Corrupted.jpg");
      Calendar dateCreated = Calendar.getInstance();
      dateCreated.setTime(new Date(otherFormatFile.lastModified()));
      metaDataIn.setCreationDate(dateCreated);
    }
    return metaDataIn;
  }


  public String renderRaw(File otherFormatFile, File jpegFileToCreate, File ppmThumbNail) throws Exception {
    String errorMessage= "Invalid RAW file: "+otherFormatFile.getAbsolutePath();
    
    dcRawConvertCmd.set(this.dcRawThumbCmdFilePos, otherFormatFile.getAbsolutePath());
    File ppmThumbNailRaw=getThumbFile(otherFormatFile, ".ppm");
    new DcRawProcessWrapper(logger, this.pluginDirectory, dcRawConvertCmd, ppmThumbNail);
    if(ppmThumbNail.exists()) {
      cjpegCmd.set(1, ppmThumbNail.getAbsolutePath());  
      new DcRawProcessWrapper(logger, this.pluginDirectory, cjpegCmd, jpegFileToCreate);
      ppmThumbNail.delete();  
      if(jpegFileToCreate.exists()) {
        if(jpegFileToCreate.length()>0) {
          errorMessage=null;              
          }
        } 
      }
    return errorMessage;
  }
  
  private void scanMetaData(File otherFormatFile, String metadata, CcXMPMetaData metaDataIn) throws IOException {
    try {
      metaDataIn.setTitle(otherFormatFile.getName());
      }
    catch (XMPException xmpe) {
      logger.LogErrorMessage(xmpe.getLocalizedMessage());
      }
    
    if (metadata != null) {
      try (BufferedReader reader = new BufferedReader(new StringReader(metadata))) {
        String line = reader.readLine();
        while (line != null) {
          try {
              scanMetaLine(line, metaDataIn);
              }
          catch (XMPException e) {
              logger.LogErrorMessage(e.getLocalizedMessage());
              }
          line = reader.readLine();
        }
      }
    }
  }
  

  private void scanMetaLine(String line, CcXMPMetaData metaDataIn) throws XMPException {
    
    
    if(line.startsWith("Timestamp: ")) {
      String ds="";  
      try {
        ds=line.substring("Timestamp: ".length());
        Date cdate = CTIME_DATE_FORMAT.parse(ds);
        Calendar calendar = Calendar.getInstance();
        calendar.setTime(cdate);        
        metaDataIn.setCreationDate(calendar);
        }
      catch (ParseException e) {
        logger.LogErrorMessage("DcRawPlugin invalid date: "+ds);
        }
      
      
      return;
      }
    if(line.startsWith("Camera: ")) {
      metaDataIn.setModel(line.substring("Camera: ".length()));
      return;
      }
    if(line.startsWith("ISO speed: ")) {
      metaDataIn.setIsoSpeed(line.substring("ISO speed: ".length()));
      return;
      }
    if(line.startsWith("Shutter: ")) {
      metaDataIn.setExposoreTime(line.substring("Shutter: ".length()));
      return;
      }
    if(line.startsWith("Aperture: ")) {
      metaDataIn.setExifApertureValue(line.substring("Aperture: ".length()));
      return;
      }
    if(line.startsWith("Focal length: ")) {
      metaDataIn.setFocalLength(line.substring("Focal length: ".length()));
      return;
      }
  }


  private File getThumbFile(File otherFormatFile, String extension) {
    String fname = otherFormatFile.getAbsolutePath();
    int p=fname.lastIndexOf(".");
    
    if(p<0) {
      return(new File(fname+".thumb"+extension));
    }
   return(new File(fname.substring(0,p)+".thumb"+extension));
  }

  
  private void copyDummySubstitute(File target, String resourceName) throws Exception {
    try(InputStream resourceAsStream = DcRawPlugin.class.getClassLoader().getResourceAsStream("de/contecon/picapport/plugin/dcraw/images/" + resourceName)) {
        Files.copy(resourceAsStream, target.toPath(), StandardCopyOption.REPLACE_EXISTING);
    } 
  }

    
}
    
    
    
    

  
 
